@@ -1,12 +1,19 @@
 {-# LANGUAGE CPP #-}
-
-#if __GLASGOW_HASKELL__ >= 708
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE EmptyDataDecls #-}
+{-# LANGUAGE TypeFamilies #-}
+
+#if __GLASGOW_HASKELL__ >= 702
+{-# LANGUAGE DeriveGeneric #-}
+#endif
+
+#if __GLASGOW_HASKELL__ >= 706
+{-# LANGUAGE PolyKinds #-}
 #endif
 
 -----------------------------------------------------------------------------
 -- |
--- Copyright   :  (C) 2008-2015 Edward Kmett
+-- Copyright   :  (C) 2008-2016 Edward Kmett
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  :  Edward Kmett <ekmett@gmail.com>
@@ -38,17 +45,48 @@ import Data.Traversable
 import Data.Typeable
 #endif
 
+#if __GLASGOW_HASKELL__ >= 702
+import GHC.Generics
+#endif
+
 -- | Make a 'Functor' over the first argument of a 'Bifunctor'.
 --
 -- Mnemonic: C__l__owns to the __l__eft (parameter of the Bifunctor),
 --           joke__r__s to the __r__ight.
 newtype Clown f a b = Clown { runClown :: f a }
   deriving ( Eq, Ord, Show, Read
+#if __GLASGOW_HASKELL__ >= 702
+           , Generic
+#endif
 #if __GLASGOW_HASKELL__ >= 708
+           , Generic1
            , Typeable
 #endif
            )
 
+#if __GLASGOW_HASKELL__ >= 702 && __GLASGOW_HASKELL__ < 708
+data ClownMetaData
+data ClownMetaCons
+data ClownMetaSel
+
+instance Datatype ClownMetaData where
+    datatypeName _ = "Clown"
+    moduleName _ = "Data.Bifunctor.Clown"
+
+instance Constructor ClownMetaCons where
+    conName _ = "Clown"
+    conIsRecord _ = True
+
+instance Selector ClownMetaSel where
+    selName _ = "runClown"
+
+instance Generic1 (Clown f a) where
+    type Rep1 (Clown f a) = D1 ClownMetaData (C1 ClownMetaCons
+        (S1 ClownMetaSel (Rec0 (f a))))
+    from1 = M1 . M1 . M1 . K1 . runClown
+    to1 = Clown . unK1 . unM1 . unM1 . unM1
+#endif
+
 instance Functor f => Bifunctor (Clown f) where
   first f = Clown . fmap f . runClown
   {-# INLINE first #-}
