@@ -1,11 +1,19 @@
 {-# LANGUAGE CPP #-}
-
-#if __GLASGOW_HASKELL__ >= 708
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE EmptyDataDecls #-}
+{-# LANGUAGE TypeFamilies #-}
+
+#if __GLASGOW_HASKELL__ >= 702
+{-# LANGUAGE DeriveGeneric #-}
 #endif
+
+#if __GLASGOW_HASKELL__ >= 706
+{-# LANGUAGE PolyKinds #-}
+#endif
+
 -----------------------------------------------------------------------------
 -- |
--- Copyright   :  (C) 2008-2015 Edward Kmett
+-- Copyright   :  (C) 2008-2016 Edward Kmett
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  :  Edward Kmett <ekmett@gmail.com>
@@ -36,17 +44,48 @@ import Data.Traversable
 import Data.Typeable
 #endif
 
+#if __GLASGOW_HASKELL__ >= 702
+import GHC.Generics
+#endif
+
 -- | Make a 'Functor' over the second argument of a 'Bifunctor'.
 --
 -- Mnemonic: C__l__owns to the __l__eft (parameter of the Bifunctor),
 --           joke__r__s to the __r__ight.
 newtype Joker g a b = Joker { runJoker :: g b }
   deriving ( Eq, Ord, Show, Read
+#if __GLASGOW_HASKELL__ >= 702
+           , Generic
+#endif
 #if __GLASGOW_HASKELL__ >= 708
+           , Generic1
            , Typeable
 #endif
            )
 
+#if __GLASGOW_HASKELL__ >= 702 && __GLASGOW_HASKELL__ < 708
+data JokerMetaData
+data JokerMetaCons
+data JokerMetaSel
+
+instance Datatype JokerMetaData where
+    datatypeName _ = "Joker"
+    moduleName _ = "Data.Bifunctor.Joker"
+
+instance Constructor JokerMetaCons where
+    conName _ = "Joker"
+    conIsRecord _ = True
+
+instance Selector JokerMetaSel where
+    selName _ = "runJoker"
+
+instance Generic1 (Joker g a) where
+    type Rep1 (Joker g a) = D1 JokerMetaData (C1 JokerMetaCons
+        (S1 JokerMetaSel (Rec1 g)))
+    from1 = M1 . M1 . M1 . Rec1 . runJoker
+    to1 = Joker . unRec1 . unM1 . unM1 . unM1
+#endif
+
 instance Functor g => Bifunctor (Joker g) where
   first _ = Joker . runJoker
   {-# INLINE first #-}
