@@ -1,12 +1,19 @@
 {-# LANGUAGE CPP #-}
-
-#if __GLASGOW_HASKELL__ >= 708
 {-# LANGUAGE DeriveDataTypeable #-}
+{-# LANGUAGE EmptyDataDecls #-}
+{-# LANGUAGE TypeFamilies #-}
+
+#if __GLASGOW_HASKELL__ >= 702
+{-# LANGUAGE DeriveGeneric #-}
+#endif
+
+#if __GLASGOW_HASKELL__ >= 706
+{-# LANGUAGE PolyKinds #-}
 #endif
 
 -----------------------------------------------------------------------------
 -- |
--- Copyright   :  (C) 2008-2015 Edward Kmett
+-- Copyright   :  (C) 2008-2016 Edward Kmett
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  :  Edward Kmett <ekmett@gmail.com>
@@ -36,14 +43,46 @@ import Data.Traversable
 import Data.Typeable
 #endif
 
+#if __GLASGOW_HASKELL__ >= 702
+import GHC.Generics
+#endif
+
 -- | Make a 'Functor' over the second argument of a 'Bifunctor'.
 newtype WrappedBifunctor p a b = WrapBifunctor { unwrapBifunctor :: p a b }
   deriving ( Eq, Ord, Show, Read
+#if __GLASGOW_HASKELL__ >= 702
+           , Generic
+#endif
 #if __GLASGOW_HASKELL__ >= 708
+           , Generic1
            , Typeable
 #endif
            )
 
+#if __GLASGOW_HASKELL__ >= 702 && __GLASGOW_HASKELL__ < 708
+data WrappedBifunctorMetaData
+data WrappedBifunctorMetaCons
+data WrappedBifunctorMetaSel
+
+instance Datatype WrappedBifunctorMetaData where
+    datatypeName = const "WrappedBifunctor"
+    moduleName = const "Data.Bifunctor.Wrapped"
+
+instance Constructor WrappedBifunctorMetaCons where
+    conName = const "WrapBifunctor"
+    conIsRecord = const True
+
+instance Selector WrappedBifunctorMetaSel where
+    selName = const "unwrapBifunctor"
+
+instance Generic1 (WrappedBifunctor p a) where
+    type Rep1 (WrappedBifunctor p a) = D1 WrappedBifunctorMetaData
+        (C1 WrappedBifunctorMetaCons
+            (S1 WrappedBifunctorMetaSel (Rec1 (p a))))
+    from1 = M1 . M1 . M1 . Rec1 . unwrapBifunctor
+    to1 = WrapBifunctor . unRec1 . unM1 . unM1 . unM1
+#endif
+
 instance Bifunctor p => Bifunctor (WrappedBifunctor p) where
   first f = WrapBifunctor . first f . unwrapBifunctor
   {-# INLINE first #-}
